@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997, 2005, 2006, 2007 Sergey Poznyakoff
+   Copyright (C) 1997, 2005, 2006, 2007, 2009 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -43,7 +43,6 @@ int parmdcl(Ident*);
 int dirdcl(Ident*);
 void skip_struct();
 Symbol *get_symbol(char *name);
-Symbol *install_ident(char *name);
 void maybe_parm_list(int *parm_cnt_return);
     
 void call(char*, int);
@@ -520,6 +519,8 @@ parse_variable_declaration(Ident *ident, int parm)
 	  }
 	  putback();
 	  skip_struct();
+	  while (tok.type == MODIFIER)
+	       nexttoken();
 	  if (tok.type == IDENTIFIER) {
 	       TOKSTK hold = tok;
 	       restore(sp);
@@ -956,9 +957,7 @@ declare(Ident *ident, int maybe_knr)
      
      if (ident->storage == AutoStorage) {
 	  undo_save_stack();
-	  sp = install(ident->name);
-	  sp->type = SymIdentifier;
-	  sp->storage = ident->storage;
+	  sp = install_ident(ident->name, ident->storage);
 	  if (parm_level) {
 	       sp->level = parm_level;
 	       sp->flag = symbol_parm;
@@ -981,7 +980,7 @@ declare(Ident *ident, int maybe_knr)
      if (sp->source) {
 	  if (ident->storage == StaticStorage
 	      && (sp->storage != StaticStorage || level > 0)) {
-	       sp = install_ident(ident->name);
+	       sp = install_ident(ident->name, ident->storage);
 	  } else {
 	       error_at_line(0, 0, filename, ident->line, 
 			     _("%s/%d redefined"),
@@ -993,8 +992,9 @@ declare(Ident *ident, int maybe_knr)
 
      sp->type = SymIdentifier;
      sp->arity = ident->parmcnt;
-     sp->storage = (ident->storage == ExplicitExternStorage) ?
-	  ExternStorage : ident->storage;
+     ident_change_storage(sp, 
+			  (ident->storage == ExplicitExternStorage) ?
+			  ExternStorage : ident->storage);
      sp->decl = finish_save_stack(ident->name);
      sp->source = filename;
      sp->def_line = ident->line;
@@ -1031,30 +1031,12 @@ declare_type(Ident *ident)
 		 ident->name);
 }
 
-Symbol *
-install_ident(char *name)
-{
-     Symbol *sp;
-
-     sp = install(name);
-     sp->type = SymIdentifier;
-     sp->arity = -1;
-     sp->storage = ExternStorage;
-     sp->decl = NULL;
-     sp->source = NULL;
-     sp->def_line = -1;
-     sp->ref_line = NULL;
-     sp->caller = sp->callee = NULL;
-     sp->level = -1;
-     return sp;
-}
-
 Symbol *
 get_symbol(char *name)
 {
-     Symbol *sp;
+     Symbol *sp = lookup(name) ;
      
-     if (sp = lookup(name)) {
+     if (sp) {
 	  for (; sp; sp = sp->next) {
 	       if (sp->type == SymIdentifier && strcmp(sp->name, name) == 0)
 		    break;
@@ -1062,7 +1044,7 @@ get_symbol(char *name)
 	  if (sp)
 	       return sp;
      }
-     return install_ident(name);
+     return install_ident(name, ExternStorage);
 }
 
 Symbol *
@@ -1077,7 +1059,9 @@ add_reference(char *name, int line)
      refptr = xmalloc(sizeof(*refptr));
      refptr->source = filename;
      refptr->line = line;
-     append_to_list(&sp->ref_line, refptr);
+     if (!sp->ref_line)
+	  sp->ref_line = linked_list_create(free);
+     linked_list_append(&sp->ref_line, refptr);
      return sp;
 }
 
@@ -1093,10 +1077,10 @@ call(char *name, int line)
      if (sp->arity < 0)
 	  sp->arity = 0;
      if (caller) {
-	  if (!symbol_in_list(caller, sp->caller))
-	       append_to_list(&sp->caller, caller);
-	  if (!symbol_in_list(sp, caller->callee))
-	       append_to_list(&caller->callee, sp);
+	  if (!data_in_list(caller, sp->caller))
+	       linked_list_append(&sp->caller, caller);
+	  if (!data_in_list(sp, caller->callee))
+	       linked_list_append(&caller->callee, sp);
      }
 }
 
@@ -1107,10 +1091,10 @@ reference(char *name, int line)
      if (!sp)
 	  return;
      if (caller) {
-	  if (!symbol_in_list(caller, sp->caller))
-	       append_to_list(&sp->caller, caller);
-	  if (!symbol_in_list(sp, caller->callee))
-	       append_to_list(&caller->callee, sp);
+	  if (!data_in_list(caller, sp->caller))
+	       linked_list_append(&sp->caller, caller);
+	  if (!data_in_list(sp, caller->callee))
+	       linked_list_append(&caller->callee, sp);
      }
 }
 
