@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997, 2005, 2006, 2007 Sergey Poznyakoff
+   Copyright (C) 1997, 2005, 2006, 2007, 2009 Sergey Poznyakoff
 
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,6 +22,22 @@
 
 static Hash_table *symbol_table;
 
+static struct linked_list *static_symbol_list;
+static struct linked_list *auto_symbol_list;
+
+static void
+append_symbol(struct linked_list **plist, Symbol *sp)
+{
+     if (sp->entry) {
+	  linked_list_unlink(sp->entry->list, sp->entry);
+	  sp->entry = NULL;
+     }
+     if (!data_in_list(sp, *plist)) {
+	  linked_list_append(plist, sp);
+	  sp->entry = (*plist)->tail;
+     }
+}
+
 struct table_entry {
      Symbol *sym;
 };
@@ -89,150 +105,157 @@ install(char *name)
 	  if (ret->sym->type != SymUndefined) 
 	       sym->next = ret->sym;
 	  ret->sym = sym;
+	  free(tp);
      }
+     sym->owner = ret;
+     if (sym->flag == symbol_temp)
+	  append_symbol(&static_symbol_list, sym);
      return sym;
 }
 
-/* Unlink the first symbol from the table entry */
-static Symbol *
-unlink_symbol(struct table_entry *tp)
+void
+ident_change_storage(Symbol *sp, enum storage storage)
 {
-     Symbol *s = tp->sym;
-     if (s) 
-	  tp->sym = s->next;
-     return s;
+     if (sp->storage == storage)
+	  return;
+     if (sp->storage == StaticStorage)
+	  /* FIXME */;
+
+     switch (storage) {
+     case StaticStorage:
+	  append_symbol(&static_symbol_list, sp);
+	  break;
+     case AutoStorage:
+	  append_symbol(&auto_symbol_list, sp);
+	  break;
+     default:
+	  break;
+     }
+     sp->storage = storage;
+}
+
+Symbol *
+install_ident(char *name, enum storage storage)
+{
+     Symbol *sp;
+
+     sp = install(name);
+     sp->type = SymIdentifier;
+     sp->arity = -1;
+     sp->storage = ExternStorage;
+     sp->decl = NULL;
+     sp->source = NULL;
+     sp->def_line = -1;
+     sp->ref_line = NULL;
+     sp->caller = sp->callee = NULL;
+     sp->level = -1;
+     ident_change_storage(sp, storage);
+     return sp;
+}
+
+/* Unlink the symbol from the table entry */
+static void
+unlink_symbol(Symbol *sym)
+{
+     Symbol *s, *prev = NULL;
+     struct table_entry *tp = sym->owner;
+     for (s = tp->sym; s; ) {
+	  Symbol *next = s->next;
+	  if (s == sym) {
+	       if (prev)
+		    prev->next = next;
+	       else
+		    tp->sym = next;
+	       break;
+	  } else
+	       prev = s;
+	  s = next;
+     }
+	       
+     sym->owner = NULL;
 }     
 
 /* Unlink and free the first symbol from the table entry */
 static void
-delete_symbol(struct table_entry *tp)
+delete_symbol(Symbol *sym)
 {
-     Symbol *s = unlink_symbol(tp);
+     unlink_symbol(sym);
      /* The symbol could have been referenced even if it is static
 	in -i^s mode. See tests/static.at for details. */
-     if (s->ref_line == NULL) 
-	  free(s);
+     if (sym->ref_line == NULL) {
+	  linked_list_destroy(&sym->ref_line);
+	  linked_list_destroy(&sym->caller);
+	  linked_list_destroy(&sym->callee);
+	  free(sym);
+     }
 }     
 
-static void cleanup_symbol_refs(Symbol *sym);
-
 /* Delete from the symbol table all static symbols defined in the current
    source.
    If the user requested static symbols in the listing, the symbol is
    not deleted, as it may have been referenced by other symbols. Instead,
-   it is unlinked from its table entry and cleanup_symbol_refs() is
-   called, so that its reference tables become usable.
+   it is unlinked from its table entry.
    NOTE: This takes advantage of the fact that install() uses LIFO strategy,
    so we don't have to check the name of the source where the symbol was
    defined. */
 
-static bool
-static_processor(void *data, void *proc_data)
+static void
+static_free(void *data)
 {
-     struct table_entry *t = data;
+     Symbol *sym = data;
+     struct table_entry *t = sym->owner;
 
-     if (t->sym
-	 && t->sym->type == SymIdentifier
-	 && t->sym->storage == StaticStorage) {
-	  if (globals_only()) 
-	       delete_symbol(t);
-	  else
-	       cleanup_symbol_refs(unlink_symbol(t));
-     }
-     return true;
-}
-
-static bool
-temp_processor(void *data, void *proc_data)
-{
-     struct table_entry *t = data;
-     
-     if (t->sym && t->sym->flag == symbol_temp) 
-	  delete_symbol(t);
-     return true;
+     if (!t)
+	  return;
+     if (sym->flag == symbol_temp
+	 || globals_only()) 
+	  delete_symbol(sym);
+     else
+	  unlink_symbol(sym);
 }
 
 void
 delete_statics()
 {
-     hash_do_for_each (symbol_table, static_processor, NULL);
-     hash_do_for_each (symbol_table, temp_processor, NULL);
+     if (static_symbol_list) {
+	  static_symbol_list->free_data = static_free;
+	  linked_list_destroy(&static_symbol_list);
+     }
 }
 
 /* See NOTE above */
-bool
-auto_processor(void *data, void *proc_data)
-{
-     struct table_entry *t = data;
-     Symbol *s = t->sym;
-     if (s) {
-	  int *level = proc_data;
-	  if (s->type == SymIdentifier && s->level == *level) {
-	       switch (s->storage) {
-	       case AutoStorage:
-		    delete_symbol(t);
-		    break;
-		    
-	       case StaticStorage:
-		    cleanup_symbol_refs(unlink_symbol(t));
-		    break;
-		    
-	       default:
-		    break;
-	       }
-	  }
-     }
-     return true;
-}
 
 /* Delete from the symbol table all auto variables with given nesting
    level. */
+int
+delete_level_autos(void *data, void *call_data)
+{
+     int level = *(int*)call_data;
+     Symbol *s = data;
+     if (s->level == level) {
+	  delete_symbol(s);
+	  return 1;
+     }
+     return 0;
+}
+
+int
+delete_level_statics(void *data, void *call_data)
+{
+     int level = *(int*)call_data;
+     Symbol *s = data;
+     if (s->level == level) {
+	  unlink_symbol(s);
+	  return 1;
+     }
+     return 0;
+}
+
 void
 delete_autos(int level)
 {
-     hash_do_for_each (symbol_table, auto_processor, &level);
-}
-
-
-/* Make all list pointers of the SYM ready for final processing.
- * This means for each list replace its entry point with its CAR
- * and throw away the first cons. The first cons holds pointers
- * to the head and tail of the list and is used to speed up appends.
- *  
- * TODO: The memory is not reclaimed
- */
-static void
-cleanup_symbol_refs(Symbol *sym)
-{
-     if (sym && sym->type == SymIdentifier) {
-	  if (sym->ref_line)
-	       sym->ref_line = CAR(sym->ref_line);
-	  if (sym->caller)
-	       sym->caller = CAR(sym->caller);
-	  if (sym->callee)
-	       sym->callee = CAR(sym->callee);
-     }
-}
-
-static bool
-cleanup_processor(void *data, void *proc_data)
-{
-     struct table_entry *t = data;
-     Symbol *sym;
-     for (sym = t->sym; sym; sym = sym->next) {
-	  cleanup_symbol_refs(sym);
-     }
-     return true;
-}
-
-
-/* Clean up all symbols from the auxiliary information.
- * See the comment for cleanup_symbol() above
- */
-void
-cleanup()
-{
-     hash_do_for_each (symbol_table, cleanup_processor, NULL);
+     linked_list_iterate(&auto_symbol_list, delete_level_autos, &level);
+     linked_list_iterate(&static_symbol_list, delete_level_statics, &level);
 }
 
 struct collect_data {
@@ -278,41 +301,28 @@ collect_symbols(Symbol ***return_sym, int (*sel)(Symbol *p))
 
 /* Special handling for function parameters */
 
-static bool
-delete_parm_processor(void *data, void *proc_data)
+int
+delete_parms_itr(void *data, void *call_data)
 {
-     struct table_entry *t = data;
-     Symbol *s = t->sym;
-     if (s) {
-	  int *level = proc_data;
-	  if (s->type == SymIdentifier && s->storage == AutoStorage
-	      && s->flag == symbol_parm && s->level > *level)
-	       delete_symbol(t);
+     int level = *(int*)call_data;
+     Symbol *s = data;
+     struct table_entry *t = s->owner;
+	  
+     if (!t)
+	  return 1;
+     if (s->type == SymIdentifier && s->storage == AutoStorage
+	 && s->flag == symbol_parm && s->level > level) {
+	  delete_symbol(s);
+	  return 1;
      }
-     return true;
+     return 0;
 }
 
 /* Delete all parameters with parameter nesting level greater than LEVEL */
 void
 delete_parms(int level)
 {
-     hash_do_for_each (symbol_table, delete_parm_processor, &level);
-}
-
-static bool
-move_parm_processor(void *data, void *proc_data)
-{
-     struct table_entry *t = data;
-     Symbol *s = t->sym;
-     if (s) {
-	  int level = *(int*)proc_data;
-	  if (s->type == SymIdentifier && s->storage == AutoStorage
-	      && s->flag == symbol_parm) {
-	       s->level = level;
-	       s->flag = symbol_none;
-	  }
-     }
-     return true;
+     linked_list_iterate(&auto_symbol_list, delete_parms_itr, &level);
 }
 
 /* Redeclare all saved parameters as automatic variables with the
@@ -320,101 +330,16 @@ move_parm_processor(void *data, void *proc_data)
 void
 move_parms(int level)
 {
-     hash_do_for_each (symbol_table, move_parm_processor, &level);
-}
+     struct linked_list_entry *p;
 
-
-typedef struct bucket Bucket;
-struct bucket {
-     Bucket *next; /* Next bucket */
-     int free;
-     Cons cons[1];
-};
+     for (p = linked_list_head(auto_symbol_list); p; p = p->next) {
+	  Symbol *s = p->data;
 
-static int bucket_nodes = 512;
-static Bucket *root_bucket, *last_bucket;
-
-void
-alloc_new_bucket()
-{
-     Bucket *bp;
-     
-     bp = malloc(sizeof(*bp) + sizeof(Cons)*(bucket_nodes-1));
-     if (!bp)
-	  return;
-     bp->next = NULL;
-     bp->free = 0;
-     if (!root_bucket) 
-	  root_bucket = last_bucket = bp;
-     else {
-	  last_bucket->next = bp;
-	  last_bucket = bp;
-     }
-}
-
-Consptr
-alloc_cons_from_bucket()
-{
-     if (!last_bucket || last_bucket->free == bucket_nodes)
-	  return NULL;
-     return &last_bucket->cons[last_bucket->free++];
-}
-
-Consptr
-alloc_cons()
-{
-     Consptr cp;
-     
-     cp = alloc_cons_from_bucket();
-     if (!cp) {
-	  alloc_new_bucket();
-	  if ((cp = alloc_cons_from_bucket()) == NULL) {
-	       error(2, 0, _("not enough core"));
+	  if (s->type == SymIdentifier && s->storage == AutoStorage
+	      && s->flag == symbol_parm) {
+	       s->level = level;
+	       s->flag = symbol_none;
 	  }
      }
-     CAR(cp) = CDR(cp) = NULL;
-     return cp;
 }
 
-/* Append a new cons to the tail of the list
- * ROOT_PTR points to a `root cons'. 
- * CAR is the car value of the cons to be created.
- *  
- * Note: Car of the root cons points to the head of the list,
- * cdr of root cons points to  the tail of the list.
- */
-Consptr
-append_to_list(Consptr *root_ptr, void *car)
-{
-     Consptr root, cons;
-     
-     if (!*root_ptr) {
-	  *root_ptr = alloc_cons();
-	  /* both car and cdr are guaranteed to be NULL */ 
-     }
-     root = *root_ptr;
-     
-     cons = alloc_cons();
-     if (!CAR(root))
-	  CAR(root) = cons;
-     
-     /* Maintain linked list */
-     if (CDR(root))
-	  CDR(CDR(root)) = cons;
-     CDR(root) = cons;
-     CAR(cons) = car;
-     return cons;
-}
-
-int
-symbol_in_list(Symbol *sym, Consptr list)
-{
-     Consptr cons;
-     
-     if (!list)
-	  return 0;
-     for (cons = CAR(list); cons; cons = CDR(cons))
-	  if ((Symbol*)CAR(cons) == sym)
-	       return 1;
-     return 0;
-}
